import pika
import pika.frame as pika_frame
import threading
from logging import Logger
from typing import Final
from pika.channel import Channel
from pika.connection import Connection
from pika.exchange_type import ExchangeType
from pika.spec import BasicProperties

MQP_CONNECTION_OPEN: Final[int] = 1
MQP_CONNECTION_CLOSED: Final[int] = 2
MQP_CONNECTION_ERROR: Final[int] = -1
MQP_INITIALIZING: Final[int] = 0


class _MqPublisher(threading.Thread):
    """
    A wrapper on *pika*, an implementation of the publisher side of a *RabbitMQ* client.

    This is an example of a publisher that can handle unexpected interactions with *RabbitMQ*,
    such as channel closing and connection interruptions.

    If *RabbitMQ* closes the connection, this implementation will attempt to reopen it.
    In these case, the execution log should be examined, as there are only a few reasons
     to justify the closing of the connection, and they are usually related to permissions or
    time limitations on the use of sockets.

    This implementation uses delivery confirmation, by providing a means to follow up
    on message delivery, and verify if such deliveries were confirme by *RabbitMQ*.
    """

    def __init__(self, mq_url: str, exchange_name: str, exchange_type: str,
                 max_reconnect_delay: int, logger: Logger = None) -> None:
        """
        Create a new instance of the publisher, with the arguments to allow it to interact with *RabbitMQ*.

        :param mq_url: A URL usada for the connection
        :param exchange_name: name of the exchange to use
        :param exchange_type: type of the exchange
        :param max_reconnect_delay: maximum delay for re-establishing lost connections, in soeconds
        :param logger: optional logger
        """
        threading.Thread.__init__(self)

        exch_type: ExchangeType | str
        match exchange_type:
            case "direct":
                exch_type = ExchangeType.direct
            case "fanout":
                exch_type = ExchangeType.fanout
            case "headers":
                exch_type = ExchangeType.headers
            case _:  # 'topic'
                exch_type = ExchangeType.topic

        # initialize instance attributes
        self.exchange_name: str = exchange_name
        self.exchange_type: ExchangeType = exch_type

        self.started_publishing: bool = False
        self.mq_url: str = mq_url
        self.stopped: bool = False
        self.acked: int = 0
        self.nacked: int = 0
        self.msg_last_filed: int = 0
        self.msg_last_sent: int = 0
        self.logger: Logger = logger
        self.publish_interval: int = 1

        self.reconnect_delay: int = 0
        self.max_reconnect_delay: int = max_reconnect_delay

        self.conn: Connection | None = None
        self.channel: Channel | None = None

        self.state: int = MQP_INITIALIZING
        self.state_msg: str = "Attenpting to instantiate the publisher"

        # structure ('n' is the sequential message number, int > 0):
        # <{ n: { "header": <str>,
        #        "body":  <bytes>,    # noqa: ERA001
        #        "mimetype": <str>,   # noqa: ERA001
        #        "routing_key": <str>
        #      },...
        # }>
        self.messages: dict | None = None

        self.messages = None
        # mutex for controlling concurrent access to the message structure
        self.msg_lock: threading.Lock = threading.Lock()

        if self.logger:
            self.logger.info("Publisher instantiated, with exchange "
                             f"'{exchange_name}' of type '{exchange_type}'")

    # ponto de entrada para a thread
    def run(self) -> None:
        """
        Initialize the publisher, by connecting with *RabbitMQ* e initiating the *IOLoop*.
        """
        # stay in the loop, until 'stop()' is invoked
        while not self.stopped:
            if self.logger:
                self.logger.info("Started")
            self.messages = {}
            self.acked = 0
            self.nacked = 0
            self.msg_last_sent = 0

            # conect with RabbitMQ
            self.conn = self.connect()

            # initiate the IOLoop, blocking until it is interrupted
            self.conn.ioloop.start()

        if self.logger:
            self.logger.info("Finished")

    def connect(self) -> Connection:
        """
        Connect with *RabbitMQ*, and return the connection identifier.

        When the connection is established, *on_connection_open* will be invooked by *pika*.

        :return: the connection obtained
        """
        if self.logger:
            # do not write user and password from URL in the log
            #   url: <protocol>//<user>:<password>@<ip-address>
            first: int = self.mq_url.find("//")
            last = self.mq_url.find("@")
            if self.logger:
                self.logger.info(f"Connecting with '{self.mq_url[0:first]}{self.mq_url[last:]}'")

        # obtain anf return the connection
        return pika.SelectConnection(
            pika.URLParameters(self.mq_url),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed)

    def on_connection_open(self, _connection: Connection) -> None:
        """
        Account for *pika*'s *callback* invocation, when the connection with *RabbitMQ* is established.

        The identifier for the object is given, to be used as needed. At the moment, it is marked as not used.

        :param _connection: the connection with RabbitMQ
        """
        self.state = MQP_CONNECTION_OPEN
        self.state_msg = "Connection was open"
        if self.logger:
            self.logger.info(self.state_msg)
        self.open_channel()

    def on_connection_open_error(self, _connection: Connection, error: Exception) -> None:
        """
         Account for *pika*'s *callback* invocation, if the connection with *RabbitMQ* could not be established.

        :param _connection: the attempted connection RabbitMQ
        :param error: the corresponding error message
        """
        self.state = MQP_CONNECTION_ERROR
        self.state_msg = f"Error establishing connection: {error}"
        delay: int = self.__get_reconnect_delay()
        if self.logger:
            self.logger.error(self.state_msg)
            self.logger.info(f"Reconnecting in {delay} seconds")
        self.conn.ioloop.call_later(delay, self.conn.ioloop.stop)

    def on_connection_closed(self, _connection: Connection, reason: Exception) -> None:
        """
        Account for *pika*'s *callback* invocation, when the connection with *RabbitMQ* is closed unexpectedly.

        In this situation, reconnecting with *RabbitMQ* is attempted.

        :param _connection: the closed connection
        :param reason: exception indicating the reason for the connection loss
        """
        self.state = MQP_CONNECTION_CLOSED
        self.state_msg = f"Connection was closed: {reason}"
        self.channel = None
        if self.stopped:
            self.conn.ioloop.stop()
        else:
            delay: int = self.__get_reconnect_delay()
            if self.logger:
                self.logger.warning(self.state_msg)
                self.logger.info(f"Reconnecting in {delay} seconds")
            self.conn.ioloop.call_later(delay, self.conn.ioloop.stop)

    def open_channel(self) -> None:
        """
        Open a new channel with *RabbitMQ*, by means of the RPC *Channel.Open* command.

        When the channel open response from *RabbitMQ* is received, the indicated *callback* is invoked by *pika*.
        """
        if self.logger:
            self.logger.info("Criando um novo canal")
        self.conn.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel: Channel) -> None:
        """
         Account for *pika*'s *callback* invocation, when the channel is open.

         The channel object isgiven as parameter, to be used as needed. With the channel now open,
         the exchange to use is declared.

        :param channel: O canal que foi aberto
        """
        if self.logger:
            self.logger.info("The channel is open. Establishing thew callback for channel closing")
        self.channel = channel
        self.channel.add_on_close_callback(self.on_channel_closed)
        self.setup_exchange()

    def on_channel_closed(self, channel: Channel, reason: Exception) -> None:
        """
        Account for *pika*'s *callback* invocation, when *RabbitMQ" unexpectedly close the channel.

        Channels are usually closed when a protocol violation is attempted,
        such as redeclaring the exchange or queue with different parameters.
        In this situation, the connection is also closed, to allow for the object *shutdown*.

        :param channel: the closed channel
        :param reason: the reason for closing the channel
        """
        if self.logger:
            self.logger.warning(f"O canal '{channel}' foi fechado: {reason}")
        self.channel = None
        if not self.stopped and not self.conn.is_closed and not self.conn.is_closing:
            self.conn.close()

    def setup_exchange(self) -> None:
        """
        Verify if the exchange is properly configured in *RabbitMQ*.

        This is done with theRPC *Exchange.Declare* command, with the parameter *passive=True*.
         If this configuraation is confirmed, *on_exchange_declare_ok* will be invoked by *pika*.
        """
        if self.logger:
            self.logger.info(f"Declarando o comutador: '{self.exchange_name}'")
        self.channel.exchange_declare(exchange=self.exchange_name,
                                      exchange_type=self.exchange_type,
                                      passive=True,
                                      durable=True,
                                      callback=self.on_exchange_declare_ok)

    def on_exchange_declare_ok(self, _unused_frame: pika_frame.Method) -> None:
        """
         Account for *pika*'s *callback* invocation, when *RabbitMQ" concludes the RPC *Exchange.Declare*.

        Enable delivery confirmations and schedule the first message to be sent to *RabbitMQ*.
        Send the RPC command *confirm_delivery* to *RabbitMQ* to enable delivery of confirmations on the channel.

        The only way to disable this is to close the channel and create a new one. When the message from *RabitMQ*
        is confirmed, the *on_delivery_confirmation* method will be invoked, passing a *Basic.Ack* or *Basic.Nack*.
        This will indicate which messages are being confirmed or rejected.

        :param _unused_frame: Exchange.DeclareOk response frame
        """
        if self.logger:
            self.logger.info(f"Comutador declarado: '{self.exchange_name}', pronto para publicação.")
        self.channel.confirm_delivery(ack_nack_callback=self.on_delivery_confirmation)
        self.started_publishing = True

    def on_delivery_confirmation(self, method_frame: pika_frame.Method) -> None:
        """
        Account for *pika*'s *callback* invocation, when *RabbitMQ* responds to a RPC *Basic.Publish* command.

        This is done with passing a *frame* *Basic.Ack* or *Basic.Nack*,
        with the delivery tag of the published message.

        The delivery tag is an integer counter, indicating the sequencial number of the message,
        sent in the channel via *Basic.Publish*. Maintenance of the message structure used for
        managing messaages, to be sent or still pending confirmation, is carried out.
        Statistics are logged.

        :param method_frame: frame Basic.Ack ou Basic.Nack
        """
        confirmation_type: str = method_frame.method.NAME.split(".")[1].lower()
        ack_multiple: bool = method_frame.method.multiple
        delivery_tag: int = method_frame.method.delivery_tag

        if self.logger:
            self.logger.info(f"Recebida confirmação de entrega: etiqueta '{delivery_tag}', "
                             f"tipo '{confirmation_type}', múltiplo: {ack_multiple}")

        if confirmation_type == "ack":
            self.acked += 1
        else:  # elif confirmation_type == "nack":
            self.nacked += 1

        with self.msg_lock:
            self.messages.pop(delivery_tag)

            if ack_multiple:
                msg_tags: list[int] = []
                for msg_tag in self.messages:
                    if msg_tag <= delivery_tag:
                        msg_tags.append(msg_tag)
                        if confirmation_type == "ack":
                            self.acked += 1
                        else:  # confirmation_type == "nack":
                            self.nacked += 1

                for msg_tag in msg_tags:
                    self.messages.pop(msg_tag)

            if self.logger:
                self.logger.info(f"Mensagens: publicadas {self.msg_last_sent}, "
                                 f"a serem confirmadas {len(self.messages)}, "
                                 f"bem sucedidas {self.acked}, mal sucedidas {self.nacked}")

    def send_message(self) -> None:
        """
        Publish a message in *RabbitMQ*, unless the publisher is stopping.

        This is done by adding the message data to the *messages* object. This object is used to verify
        the delivery confirmation in *on_delivery_confirmation*.
        """
        # o canal existe e está aberto ?
        if self.channel is not None and self.channel.is_open:
            # sim, prossiga
            self.msg_last_sent += 1

            with self.msg_lock:
                message: dict = self.messages[self.msg_last_sent]

            properties: BasicProperties = pika.BasicProperties(app_id="mq-publisher",
                                                               content_type=message["mimetype"],
                                                               headers=message["headers"])
            routing_key: str = message.get("routing_key")

            msg_body: bytes = message["body"]
            self.channel.basic_publish(exchange=self.exchange_name,
                                       routing_key=routing_key,
                                       body=msg_body,
                                       properties=properties)
            if self.logger:
                self.logger.info(f"Msg '{self.msg_last_sent}' publicada, chave '{routing_key}'")
        elif self.logger:
            # não, reporte o erro
            self.logger.error("Não é possível publicar: "
                              "não há canal aberto com o servidor de mensagens")

    def publish_message(self, errors: list[str], msg_body: str | bytes, routing_key: str,
                        msg_mimetype: str = "application/text", msg_headers: str = None) -> None:
        """
        Publish a message in *RabbitMQ*, unless the publisher is stopping.

        This is the interface for external requests for message publishing.
        *RabbitMQ* is told to invoke *send_message* in *publish_interval* seconds.
        The delivery intervals may be accelerated or decelerated, by changing this variable.
        """
        # does the channel exist and is open ?
        if self.channel is not None and self.channel.is_open:
            # yes, proceed
            msg_bytes: bytes = msg_body if isinstance(msg_body, bytes) else msg_body.encode()
            self.msg_last_filed += 1

            with self.msg_lock:
                self.messages[self.msg_last_filed] = {"headers": msg_headers,
                                                      "body": msg_bytes,
                                                      "mimetype": msg_mimetype,
                                                      "routing_key": routing_key}

            # schedule message delivery to happen in 'publish_interval' seconds
            self.conn.ioloop.call_later(self.publish_interval, self.send_message)
            if self.logger:
                self.logger.info(f"Msg '{self.msg_last_filed}' scheduled for publication in "
                                 f"{self.publish_interval}s, routing key '{routing_key}': {msg_bytes.decode()}")
        else:
            # no, report the error
            errmsg: str = "Messagen refused: no open channel to the message server exists"
            errors.append(errmsg)
            if self.logger:
                self.logger.error(errmsg)

    def get_state(self) -> int:
        """
        Return the current state of the events publisher.

         This should be one of:
            - MQP_CONNECTION_OPEN
            - MQP_CONNECTION_CLOSED
            - MQP_CONNECTION_ERROR
            - MQP_INITIALIZING

        :return: the current state of the publisher
        """
        return self.state

    def get_state_msg(self) -> str:
        """
        Return the message associated with the current state of the publisher.

        :return: the current state message.
        """

    def stop(self) -> None:
        """
        Stop the publisher, by closing the channel and the connection.

        A flag is turned on, to signal the need for interrupting the scheduling of new messages publication.
        """
        if self.logger:
            self.logger.info("Finishing...")
        self.stopped = True
        self.close_channel()
        self.close_connection()

    def close_channel(self) -> None:
        """
        Fecha o canal com *RabbitMQ*, enviando o comando RPC *Channel.Close*.
        """
        if self.channel:
            if self.logger:
                self.logger.info("Closing the channel...")
            self.channel.close()

    def close_connection(self) -> None:
        """
        Fecha a cconexão com o RabbitMQ.
        """
        if self.conn:
            if self.logger:
                self.logger.info("Closing the connection...")
            self.conn.close()

    def __get_reconnect_delay(self) -> int:
        """
        Update and return the value of the reconnection delay.

        This value is incremented by 1 every time it is retrieved, until the maximum value is reached.

        :return: the reconnection delay, in seconds.
        """
        if self.started_publishing:
            self.reconnect_delay = 0
        else:
            self.reconnect_delay += 1

        self.reconnect_delay = max(self.reconnect_delay, self.max_reconnect_delay)

        return self.reconnect_delay



import logging
import sys
import time
from pypomes_core import exc_format
from typing import Final
from .publisher_pomes import (
    MQ_CONNECTION_URL, MQ_EXCHANGE_NAME,
    MQ_EXCHANGE_TYPE, MQ_ROUTING_BASE, MQ_MAX_RECONNECT_DELAY
)
from .mq_subscriber import (
    MQS_CONNECTION_ERROR, MQS_INITIALIZING,
    _MqSubscriberMaster
)

__DEFAULT_BADGE: Final[str] = "__default__"

# dict holding the subscribers created:
#   <{ <badge-1>: <subscriber-master-instance-1>,
#     ...
#     <badge-n>: <subscriber-master-instance-n>
#   }>
__subscribers: dict = {}


def subscriber_create(errors: list[str], queue_name: str, msg_target: callable,
                      badge: str = None, max_reconnect_delay: int = MQ_MAX_RECONNECT_DELAY,
                      logger: logging.Logger = None) -> None:
    """
    Create the asynchronous subscriber.

    This is a wrapper around the package *Pika*, an implementation for a *RabbitMQ* client.

    :param errors: incidental errors
    :param queue_name: queue to use
    :param msg_target: the callback to reach the messager cosumer
    :param badge: optional badge identifying the publisher
    :param max_reconnect_delay: maximum delay for re-establishing lost connections, in seconds
    :param logger: optional logger
    :return: True if the subscriber was created, or False otherwise
    """
    # define the badge
    curr_badge: str = badge or __DEFAULT_BADGE

    # has the scheduler been instantiated ?
    if __get_subscriber(errors, curr_badge, False) is None:
        # no, instantiate it
        try:
            __subscribers[curr_badge] = _MqSubscriberMaster(MQ_CONNECTION_URL, MQ_EXCHANGE_NAME,
                                                            MQ_EXCHANGE_TYPE, f"{MQ_ROUTING_BASE}.{queue_name}",
                                                            msg_target, max_reconnect_delay, logger)
            __subscribers[curr_badge].daemon = True
        except Exception as e:
            errors.append(f"Error creating the subscriber '{badge or __DEFAULT_BADGE}': "
                          f"{exc_format(e, sys.exc_info())}")


def subscriber_destroy(badge: str = None) -> None:
    """
    Destroy the subscriber identified by *badge*. *Noop* if the subscriber does not exist.

    :param badge: optional badge identifying the scheduler
    """
    # define the badge and retrieve the corresponding subscriber
    curr_badge: str = badge or __DEFAULT_BADGE
    subscriber: _MqSubscriberMaster = __subscribers.get(curr_badge)

    # does the subscriber exist ?
    if subscriber:
        # yes, stop and discard it
        subscriber.stop()
        __subscribers.pop(curr_badge)


def subscriber_start(errors: list[str], badge: str = None) -> bool:
    """
    Start the subscriber identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :return: True if the publisher has been started, False otherwise
    """
    # initialize the return variable
    result: bool = False

    # retrieve the publisher
    subscriber: _MqSubscriberMaster = __get_subscriber(errors, badge)

    # was the subscriber retrieved ?
    if subscriber:
        # yoes, proceed
        try:
            subscriber.start()
        except Exception as e:
            errors.append(f"Error starting the subscriber '{badge or __DEFAULT_BADGE}': "
                          f"{exc_format(e, sys.exc_info())}")

        # any errors ?
        if len(errors) == 0:
            # no, wait for the conclusion
            while subscriber.consumer.get_state() == MQS_INITIALIZING:
                time.sleep(0.001)

            # did connecting with the subscriber fail ?
            if subscriber.consumer.get_state() == MQS_CONNECTION_ERROR:
                # yes, report the error
                errors.append(f"Error starting the subscriber '{badge or __DEFAULT_BADGE}': "
                              f"{subscriber.consumer.get_state_msg()}")

    return result


def subscriber_stop(errors: list[str], badge: str = None) -> bool:
    """
    Stop the subscriber identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the subscriber
    :return: True if the subscriber has been stopped, False otherwise
    """
    # initialize the return variable
    result: bool = False

    # retrieve the subscriber
    subscriber: _MqSubscriberMaster = __get_subscriber(errors, badge)

    # was the publisher retrieved ?
    if subscriber:
        # yes, proceed
        subscriber.stop()
        result = True

    return result


def subscriber_get_state(errors: list[str], badge: str = None) -> int:
    """
    Retrieve and return the current state of the subscriber identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the subscriber
    :return: the current state of the subscriber
    """
    # initialize the return variable
    result: int | None = None

    # retrieve the subscriber
    subscriber: _MqSubscriberMaster = __get_subscriber(errors, badge)

    # was the subscriber retrieved ?
    if subscriber:
        # yes, proceed
        result = subscriber.consumer.get_state()

    return result


def subscriber_get_state_msg(errors: list[str], badge: str = None) -> str:
    """
    Retrieve and return the message associated with the current state of the subscriber identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the subscriber
    :return: the message associated with the current state of the subscriber
    """
    # initialize the return variable
    result: str | None = None

    # retrieve the subscriber
    subscriber: _MqSubscriberMaster = __get_subscriber(errors, badge)

    # was the subscriber retrieved ?
    if subscriber:
        # yes, proceed
        result = subscriber.consumer.get_state_msg()

    return result


def __get_subscriber(errors: list[str], badge: str, must_exist: bool = True) -> _MqSubscriberMaster:
    """
    Retrieve the subscriber identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :param must_exist: True if publisher must exist
    :return: the publisher retrieved, or None otherwise
    """
    curr_badge = badge or __DEFAULT_BADGE
    result: _MqSubscriberMaster = __subscribers.get(curr_badge)
    if must_exist and not result:
        errors.append(f"Subscriber '{curr_badge}' has not been created")

    return result



import time
import threading
from logging import Logger
from typing import Final
from pika import frame as pika_frame
from pika import spec as pika_spec
from pika import SelectConnection, URLParameters
from pika.channel import Channel
from pika.connection import Connection
from pika.exchange_type import ExchangeType

MQS_CONNECTION_OPEN: Final[int] = 1
MQS_CONNECTION_CLOSED: Final[int] = 2
MQS_CONNECTION_ERROR: Final[int] = -1
MQS_INITIALIZING: Final[int] = 0


class _MqSubscriber:
    """
    A wrapper on *pika*, an implementation of the subscriber side of a *RabbitMQ* client.

    This is an example of a consumer that can handle unexpected interactions with *RabbitMQ*,
    such as channel closing and connection interruptions.

    If *RabbitMQ* closes the connection, this object will stop and indicate that reconnecting is required.
    In this case, the execution log should be examined, as the reasons why the connection can be
    closed are limited, and are often tied to permissions-related issues, or to socket timeouts.
    If the channel is closed, it indicates a problem with one of the commands that were issued.
    """

    def __init__(self, mq_url: str, exchange_name: str, exchange_type: str,
                 queue_name: str, msg_target: callable, logger: Logger = None) -> None:
        """
        Create an instance of the consumer, witth the arguments needed for interacting with *RabbitMQ*.

        :param mq_url: URL used in the connection
        :param exchange_name: name of the exchange to use
        :param exchange_type: type of the exchange
        :param queue_name: name of the message queue to use
        :param msg_target: callback for message deliveries
        :param logger: optional logger
        """
        exch_type:  ExchangeType
        match exchange_type:
            case "direct":
                exch_type = ExchangeType.direct
            case "fanout":
                exch_type = ExchangeType.fanout
            case "headers":
                exch_type = ExchangeType.headers
            case _:  # 'topic'
                exch_type = ExchangeType.topic

        # initialize instance attributes
        self.exchange_name = exchange_name
        self.exchange_type: ExchangeType = exch_type

        self.should_reconnect: bool = False
        self.started_consumption: bool = False
        self.logger: Logger = logger
        self.closing: bool = False
        self.consuming: bool = False
        self.consumer_tag: str | None = None
        self.mq_url: str = mq_url
        self.msg_target: callable = msg_target

        self.state: int = MQS_INITIALIZING
        self.state_msg: str = "Attenpting to initialize the subscriber"

        self.conn: Connection | None = None
        self.channel: Channel | None = None
        self.queue_name: str = queue_name

        # parameter for channel QOS - for higher yield in production, try higher values
        self.prefetch_count: int = 1

        if self.logger:
            self.logger.info("Instantiated, with exchange "
                             f"'{exchange_name}' type '{exchange_type}', and queue '{queue_name}'")

    def run(self) -> None:
        """
        Run the consumer, connecting it to *RabbitMQ*.

        The *IOLoop* is started, to block and allow the *SelectConnection* to operate.
        """
        self.conn = self.connect()
        self.conn.ioloop.start()

    def connect(self) -> Connection:
        """
        Connect with *RabbitMQ*, returning the connection identifier.

        When connection is established, *on_connection_open* will be invoke by *pika*.

        :return: the connection obtained
        """
        if self.logger:
            # supress logging user and password in the URL
            #   url: <protocol>//<user>:<password>@<ip-address>
            first: int = self.mq_url.find("//")
            last = self.mq_url.find("@")
            self.logger.info(f"Connecting with '{self.mq_url[0:first]}{self.mq_url[last:]}'")

        # obtain and return the connection
        return SelectConnection(
            parameters=URLParameters(self.mq_url),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed)

    def on_connection_open(self, _unused_connection: Connection) -> None:
        """
        Account for *pika*'s *callback* invocation, when the connection with *RabbitMQ* is established.

        The identifier for the connection object is passed as parameter, in case it is needed.
        At the moment it is marked as not used.

        :param _unused_connection: the connection with RabbitMQ
        """
        self.state = MQS_CONNECTION_OPEN
        msg: str = f"Connection established: queue '{self.queue_name}'"
        self.state_msg = msg
        if self.logger:
            self.logger.info(msg)
        self.open_channel()

    def on_connection_open_error(self, _unused_connection: Connection, error: str) -> None:
        """
        Account for *pika*'s *callback* invocation, if connecting with *RabbitMQ* cannot be accomplished.

        :param _unused_connection: the attempted connection with RabbitMQ
        :param error: the associated error message
        """
        self.state = MQS_CONNECTION_ERROR
        msg: str = f"Error attempting to connect: {error}"
        self.state_msg = msg
        if self.logger:
            self.logger.error(msg)
        self.reconnect()

    def on_connection_closed(self, _unused_connection: Connection, reason: Exception) -> None:
        """
        Account for *pika*'s *callback* invocation, when a connection with *RabbitMQ* is closed unexpectedly.

        In this situation, reconnecting with *RabbitMQ* is attempted.

        :param _unused_connection: the closed connection
        :param reason: exception indicating the reason for the connection loss
        """
        self.state = MQS_CONNECTION_CLOSED
        msg: str = f"Connection was closed: {reason}"
        self.state_msg = msg
        if self.logger:
            self.logger.warning(msg)
        self.channel = None

        if self.closing:
            self.conn.ioloop.stop()
        else:
            self.reconnect()

    def close_connection(self) -> None:
        """
        Close the connection with *RabbitMQ*.
        """
        self.consuming = False
        if self.conn.is_closing:
            if self.logger:
                self.logger.info("Connection closing...")
        elif self.conn.is_closed:
            if self.logger:
                self.logger.info("Connection already closed")
        else:
            if self.logger:
                self.logger.info("Closing the connection")
            self.conn.close()

    def reconnect(self) -> None:
        """
        Account for *pika*'s *callback* invocation, if the connection is closed, or unable to open.

        Indicate that reconnecting is needed, and then interrupt the *IOloop*.
        """
        self.should_reconnect = True
        self.stop()

    def open_channel(self) -> None:
        """
        Open a new channel with *RabbitMQ*, invoking the RPC *Channel.Open* command.

        When *RabbitMQ* responds that the channel is open, the *callback* indicated in *on_channel_open_callback*
        will be invoked by *pika*.
        """
        if self.logger:
            self.logger.info("Creating a new channel...")
        self.conn.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel: Channel) -> None:
        """
        Account for *pika*'s *callback* invocation, when the channel is open.

        A reference to the channel object is passed as parameter, in case it is needed.
        With the channel open, the exchange to be used is declared.

        :param channel: the open channel
        """
        if self.logger:
            self.logger.info("The channel is open, the callback for its closing is established")
        self.channel = channel
        self.channel.add_on_close_callback(self.on_channel_closed)
        self.setup_exchange()

    def on_channel_closed(self, channel: Channel, reason: Exception) -> None:
        """
        Account for *pika*'s *callback* invocation, when the channel is unexpectedly closed.

        Channels are usually closed swhen a protocol violation is attempted,
        such as re-declaring a switch or queue with different parameters.
        In this case, the connection is closed to allow for the shutdown of the object.

        :param channel: the closed channel
        :param reason: exception indicating the reason for the channel loss
        """
        self.logger.warning(f"The channel '{channel}' was closed: {reason}")
        self.close_connection()

    def setup_exchange(self) -> None:
        """
        Check that the exchange is configured in *RabbitMQ*.

        This is done by invoking the RPC command *Exchange.Declare* with the parameter *passive=True*.
        If this setting is confirmed, *on_exchange_declare_ok* will be invoked by *pika*.
        """
        if self.logger:
            self.logger.info(f"Declaring the exchange: '{self.exchange_name}'")
        self.channel.exchange_declare(exchange=self.exchange_name,
                                      exchange_type=self.exchange_type,
                                      passive=True,
                                      durable=True,
                                      callback=self.on_exchange_declare_ok)

    def on_exchange_declare_ok(self, _unused_frame: pika_frame.Method) -> None:
        """
        Account for *pika*'s *callback* invocation, when *RabbitMQ* concludes the RPC *Exchange.Declare* command.

        :param _unused_frame: Exchange.DeclareOk response frame
        """
        if self.logger:
            self.logger.info(f"Exchange declared: '{self.exchange_name}")
        self.setup_queue()

    def setup_queue(self) -> None:
        """
        Check that the queue is configured in *RabbitMQ*.

        This is done by invoking the RPC *Queue.Declare* command with the parameter "passive=True".
        If this setting is confirmed, *on_queue_declare_ok* will be invoked by *pika*.
        """
        if self.logger:
            self.logger.info(f"Declaring the queue '{self.queue_name}'")
        self.channel.queue_declare(queue=self.queue_name,
                                   passive=True,
                                   durable=True,
                                   callback=self.on_queue_declare_ok)

    def on_queue_declare_ok(self, _unused_frame: pika_frame.Method) -> None:
        """
        Account for *pika*'s *callback* invocation, when the RPC *Queue.Declare* call in *setup_queue* is finished.

        Here the queue is bound to the exchange with the given routing key,
        by the invocation of the RPC *Queue.Bind* command.
        When this binding is concluded, *on_bind_ok* will be invoked by *pika*.

        :param _unused_frame: the Queue.DeclareOk frame
        """
        if self.logger:
            self.logger.info(f"Exchange '{self.exchange_name}' bound ti queue '{self.queue_name}'")
        self.setup_qos()

    def setup_qos(self) -> None:
        """
        Configure the consumer prefetch.

        This is done by defining the number of messages to be delivered, with their delivery declarations pending.
        In this case, the consumer must always signal the delivery of every message to *RabbitMQ*.
        Different prefetch values may be experimented with, so that the desired performance is achieved.
        """
        self.channel.basic_qos(prefetch_count=self.prefetch_count,
                               callback=self.on_basic_qos_ok)

    def on_basic_qos_ok(self, _unused_frame: pika_frame.Method) -> None:
        """
         Account for *pika*'s *callback* invocation, when the *Basic.QoS* function is concluded.

        At this point, message consumption starts with a call to *start_consuming*,
        which will invoke the necessary RPC commands to start the process.

        :param _unused_frame: the response frame Basic.QosOk
        """
        if self.logger:
            self.logger.info(f"QOS configured to {self.prefetch_count}")
        self.start_consuming()

    def start_consuming(self) -> None:
        """
        Configure the consumer, initially by invoking *add_on_cancel_callback* in the channel.

        This is done so that the consumer be notified if *RabbitMQ* for whatever reason cancels him.
        Then the RPC *Basic.Consume* command is issued, which returns the tag used to
        uniquely identify the consumer to *RabbitMQ*. The value of this tag is retained,
        so that it can be used when canceling the consumer. The *on_message* method is passed as *callback*,
        to be invoked when new messages arrive.
        """
        if self.logger:
            self.logger.info("Adding callback for consumer cancellation")
        self.channel.add_on_cancel_callback(self.on_consumer_cancelled)

        self.consumer_tag = self.channel.basic_consume(queue=self.queue_name,
                                                       on_message_callback=self.on_message)
        self.started_consumption = True
        self.consuming = True

    def on_consumer_cancelled(self, method_frame: pika_frame.Method) -> None:
        """
        Account for *pika*'s *callback* invocation, when *RabbitMQ* sends *Basic.Cancel* to the consumer.

        :param method_frame: The Basic.Cancel frame
        """
        self.logger.info(f"The consumer has been remotely cancelled, finishing: {method_frame}")
        if self.channel:
            self.channel.close()

    def on_message(self, _unused_channel: Channel, basic_deliver: pika_spec.Basic.Deliver,
                   properties: pika_spec.BasicProperties, msg_body: bytes) -> None:
        """
        Account for *pika*'s *callback* invocation, when a message from *RabbitMQ* is delivered.

        The *basic_deliver* object that is passed contains the exchange, the routing key,
        the delivery label and a message resend flag. The *properties* parameter contains the message properties.
        The *msg_body* parameter contains the body of the sent message. Receipt of the message is then declared
        to *RabbitMQ* by sending the RPC *Basic.Ack* command with the corresponding delivery tag.

        :param _unused_channel: the Channel object
        :param basic_deliver: the Basic.Deliver object
        :param properties: the Spec.BasicProperties object
        :param msg_body: the body of the message
        """
        if self.logger:
            self.logger.info(f"Msg '{basic_deliver.delivery_tag}' "
                             f"received from '{properties.app_id}': {msg_body.decode()}")
        self.channel.basic_ack(basic_deliver.delivery_tag)

        # send message to recipient
        self.msg_target(msg_body)

    def stop_consuming(self) -> None:
        """
        Send the RPC *Basic.Cancel* command to inform *RabbitMQ* of the decision to stop consuming messages.
        """
        if self.channel:
            if self.logger:
                self.logger.info("Sending the RPC Basic.Cancel command to RabbitMQ")
            self.channel.basic_cancel(consumer_tag=self.consumer_tag,
                                      callback=self.on_cancel_ok)

    def on_cancel_ok(self, _unused_frame: pika_frame.Method) -> None:
        """
        Account for *pika*'s *callback* invocation, when *RabbitMQ* acknowledges a consumer cancellation.

        At this point the channel is closed, which will cause *on_channel_closed* to be invoked,
        which in turn will close the connection.

        :param _unused_frame: The Basic.CancelOk frame
        """
        self.consuming = False
        if self.logger:
            self.logger.info(f"RabbitMQ acknowledged the consumer cancellation: '{self.consumer_tag}'")
        self.close_channel()

    def close_channel(self) -> None:
        """
        Close the channel with *RabbitMQ* cleanly, by issuing the RPC *Channel.Close* command.
        """
        if self.logger:
            self.logger.info("Closing the channel...")
        self.channel.close()

    def get_state(self) -> int:
        """
        Return the current state of the consumer.

        The state is one of:
            - MQS_CONNECTION_OPEN
            - MQS_CONNECTION_CLOSED
            - MQS_CONNECTION_ERROR
            - MQS_INITIALIZING

        :return: the current state of the consumer.
        """
        return self.state

    def get_state_msg(self) -> str:
        """
        Return the message associated with the current state of the consumer.

        :return: the state message.
        """

    def stop(self) -> None:
        """
        Close the connection cleanly with *RabbitMQ*, stopping the consumer.

        When *RabbitMQ* confirms the cancellation, *on_cancel_ok* will be invoked by *pika*,
        which will close the channel and terminate the connection. The *IOLoop* is started again,
        because it needs to be running for *pika* to be able to communicate with *RabbitMQ*.
        All commands issued before starting *IOLoop* are stored in a buffer, but not reprocessed.
        """
        if not self.closing:
            self.closing = True
            if self.logger:
                self.logger.info("Stopping...")
            if self.consuming:
                self.stop_consuming()
                self.conn.ioloop.start()
            else:
                self.conn.ioloop.stop()
            if self.logger:
                self.logger.info("Stopped")


class _MqSubscriberMaster(threading.Thread):
    """
    Object in charge of reconnecting the consumer with *RabbitMQ*.

    This reconnection is carried out if the consumer indicates that reconnecting is necessary.
    """

    def __init__(self, mq_url: str, exchange_name: str, exchange_type: str, queue_name: str,
                 msg_target: callable, max_reconnect_delay: int, logger: Logger = None) -> None:

        threading.Thread.__init__(self)

        # initialize instance attributes
        self.mq_url: str = mq_url
        self.msg_target: callable = msg_target
        self.queue_name: str = queue_name
        self.exchange_name = exchange_name
        self.exchange_type: str = exchange_type
        self.logger: Logger = logger
        self.reconnect_delay = 0
        self.max_reconnect_delay: int = max_reconnect_delay
        self.stopped: bool = False

        # instantiate the consumer
        self.consumer: _MqSubscriber = _MqSubscriber(self.mq_url, self.exchange_name, self.exchange_type,
                                                     self.queue_name, self.msg_target, self.logger)

    def run(self) -> None:
        """
        Entry point for the thread.
        """
        while True:
            # run the consumer, blocking it until it is interrupted
            self.consumer.run()

            # stop the consumer
            self.consumer.stop()

            if self.stopped or not self.__maybe_reconnect():
                break

    def stop(self) -> None:

        self.stopped = True
        if self.consumer:
            self.consumer.stop()
            self.consumer = None

    def __maybe_reconnect(self) -> bool:
        """
        Decide whether the consumer must be recreated, so that the connection can be re-established.

        :return: the decision to re-establish the connection
        """
        result: bool = self.consumer.should_reconnect

        if result:
            reconnect_delay = self.__get_reconnect_delay()
            if self.logger:
                self.logger.info(f"Reconnecting in {reconnect_delay} seconds")
            time.sleep(reconnect_delay)

            # create a new consumer instance
            self.consumer = _MqSubscriber(self.mq_url, self.exchange_name, self.exchange_type,
                                          self.queue_name, self.msg_target, self.logger)

        return result

    def __get_reconnect_delay(self) -> int:
        """
        Update and return the value of the reconnection delay.

        This value is incremented by 1 every time it is retrieved, until the maximum value is reached.

        :return: the reconnection delay, in seconds.
        """
        if self.consumer.started_consumption:
            self.reconnect_delay = 0
        else:
            self.reconnect_delay += 1

        self.reconnect_delay = max(self.reconnect_delay, self.max_reconnect_delay)

        return self.reconnect_delay



import logging
import time
import sys
from typing import Final
from pypomes_core import APP_PREFIX, env_get_str, env_get_int, exc_format
from .mq_publisher import (
    MQP_CONNECTION_ERROR, MQP_INITIALIZING,
    _MqPublisher
)

__DEFAULT_BADGE: Final[str] = "__default__"

# environment variables
MQ_CONNECTION_URL: Final[str] = env_get_str(f"{APP_PREFIX}_MQ_CONNECTION_URL")
MQ_EXCHANGE_NAME: Final[str] = env_get_str(f"{APP_PREFIX}_MQ_EXCHANGE_NAME")
MQ_EXCHANGE_TYPE: Final[str] = env_get_str(f"{APP_PREFIX}_MQ_EXCHANGE_TYPE")
MQ_ROUTING_BASE: Final[str] = env_get_str(f"{APP_PREFIX}_MQ_ROUTING_BASE")
MQ_MAX_RECONNECT_DELAY: Final[int] = env_get_int(f"{APP_PREFIX}_MQ_MAX_RECONNECT_DELAY", 30)

# dict holding the publishers created:
#   <{ <badge-1>: <publisher-instance-1>,
#     ...
#     <badge-n>: <publisher-instance-n>
#   }>
__publishers: dict = {}


def publisher_create(errors: list[str], badge: str = None,
                     max_reconnect_delay: int = MQ_MAX_RECONNECT_DELAY,
                     logger: logging.Logger = None) -> None:
    """
    Create the threaded events publisher.

    This is a wrapper around the package *Pika*, an implementation for a *RabbitMQ* client.
    If a publisher with thw same bqadge already exists, it is not re-created.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :param max_reconnect_delay: maximum delay for re-establishing lost connections, in seconds
    :param logger: optional logger
    """
    # define the badge
    curr_badge: str = badge or __DEFAULT_BADGE

    # has the publisher been instantiated ?
    if __get_publisher(errors, curr_badge, False) is None:
        # no, instantiate it
        try:
            __publishers[curr_badge] = _MqPublisher(MQ_CONNECTION_URL, MQ_EXCHANGE_NAME,
                                                    MQ_EXCHANGE_TYPE, max_reconnect_delay, logger)
            __publishers[curr_badge].daemon = True
        except Exception as e:
            errors.append(f"Error creating the publisher '{badge or __DEFAULT_BADGE}': "
                          f"{exc_format(e, sys.exc_info())}")


def publisher_destroy(badge: str = None) -> None:
    """
    Destroy the publisher identified by *badge*. *Noop* if the publisher does not exist.

    :param badge: optional badge identifying the scheduler
    """
    # define the badge and retrieve the corresponding publisher
    curr_badge: str = badge or __DEFAULT_BADGE
    publisher: _MqPublisher = __publishers.get(curr_badge)

    # does the publisher exist ?
    if publisher:
        # yes, stop and discard it
        publisher.stop()
        __publishers.pop(curr_badge)


def publisher_start(errors: list[str], badge: str = None) -> bool:
    """
    Start the publisher identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :return: True if the publisher has been started, False otherwise
    """
    # initialize the return variable
    result: bool = False

    # retrieve the publisher
    publisher: _MqPublisher = __get_publisher(errors, badge)

    # was the publisher retrieved ?
    if publisher:
        # yes, proceed
        try:
            publisher.start()
        except Exception as e:
            errors.append(f"Error starting the publisher '{badge or __DEFAULT_BADGE}': "
                          f"{exc_format(e, sys.exc_info())}")

        # were there errors ?
        if len(errors) == 0:
            # no, wait for the conclusion
            while publisher.get_state() == MQP_INITIALIZING:
                time.sleep(0.001)

            # did connecting with the publisher fail ?
            if publisher.get_state() == MQP_CONNECTION_ERROR:
                # yes, report the error
                errors.append(f"Error starting the publisher '{badge or __DEFAULT_BADGE}': "
                              f"{publisher.get_state_msg()}")

    return result


def publisher_stop(errors: list[str], badge: str = None) -> bool:
    """
    Stop the publisher identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :return: True if the publisher has been stopped, False otherwise
    """
    # initialize the return variable
    result: bool = False

    # retrieve the publisher
    publisher: _MqPublisher = __get_publisher(errors, badge)

    # was the publisher retrieved ?
    if publisher:
        # yes, proceed
        publisher.stop()
        result = True

    return result


def publisher_get_state(errors: list[str], badge: str = None) -> int:
    """
    Retrieve and return the current state of the publisher identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :return: the current state of the publisher
    """
    # initialize the return variable
    result: int | None = None

    # retrieve the publisher
    publisher: _MqPublisher = __get_publisher(errors, badge)

    # was the publisher retrieved ?
    if publisher:
        # yes, proceed
        result = publisher.get_state()

    return result


def publisher_get_state_msg(errors: list[str], badge: str = None) -> str:
    """
    Retrieve and return the message associated with the current state of the publisher identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :return: the message associated with the current state of the publisher
    """
    # initialize the return variable
    result: str | None = None

    # retrieve the publisher
    publisher: _MqPublisher = __get_publisher(errors, badge)

    # was the publisher retrieved ?
    if publisher:
        # yes, proceed
        result = publisher.get_state_msg()

    return result


def publisher_publish(errors: list[str], msg_body: str | bytes, routing_key: str,
                      msg_mimetype: str = "application/text", msg_headers: str = None, badge: str = None) -> bool:
    """
    Send a message to the publisher identified by *badge*, for publishing.

    :param errors: incidental errors
    :param msg_body: body of the message
    :param routing_key: key for message routing
    :param msg_mimetype: message mimetype (defaults to type text)
    :param msg_headers: optional message headers
    :param badge:  optional badge identifying the publisher
    :return: True if the message was published, False otherwise
    """
    # initialize the return variable
    result: bool = False

    # retrieve the publisher
    publisher: _MqPublisher = __get_publisher(errors, badge)
    
    # was the publisher retrieved ?
    if publisher:
        # yes, proceed
        try:
            publisher.publish_message(errors, msg_body,
                                      f"{MQ_ROUTING_BASE}.{routing_key}", msg_mimetype, msg_headers)
            result = True
        except Exception as e:
            errors.append(f"Error publishing message: {exc_format(e, sys.exc_info())}")

    return result


def __get_publisher(errors: list[str], badge: str, must_exist: bool = True) -> _MqPublisher:
    """
    Retrieve the publisher identified by *badge*.

    :param errors: incidental errors
    :param badge: optional badge identifying the publisher
    :param must_exist: True if publisher must exist
    :return: the publisher retrieved, or None otherwise
    """
    curr_badge = badge or __DEFAULT_BADGE
    result: _MqPublisher = __publishers.get(curr_badge)
    if must_exist and not result:
        errors.append(f"Publisher '{curr_badge}' has not been created")

    return result



from .mq_publisher import (
    MQP_CONNECTION_OPEN, MQP_CONNECTION_CLOSED, MQP_CONNECTION_ERROR, MQP_INITIALIZING,
)
from .mq_subscriber import (
    MQS_CONNECTION_OPEN, MQS_CONNECTION_CLOSED, MQS_CONNECTION_ERROR, MQS_INITIALIZING,
)
from .publisher_pomes import (
    MQ_CONNECTION_URL, MQ_EXCHANGE_NAME,
    MQ_EXCHANGE_TYPE, MQ_ROUTING_BASE, MQ_MAX_RECONNECT_DELAY,
    publisher_create, publisher_destroy, publisher_start, publisher_stop,
    publisher_get_state, publisher_get_state_msg, publisher_publish
)
from .subscriber_pomes import (
    subscriber_create, subscriber_destroy, subscriber_start, subscriber_stop,
    subscriber_get_state, subscriber_get_state_msg,
)

__all__ = [
    # mq_publisher
    "MQP_CONNECTION_OPEN", "MQP_CONNECTION_CLOSED", "MQP_CONNECTION_ERROR", "MQP_INITIALIZING",
    # mq_subscriber
    "MQS_CONNECTION_OPEN", "MQS_CONNECTION_CLOSED", "MQS_CONNECTION_ERROR", "MQS_INITIALIZING",
    # publisher_pomes
    "MQ_CONNECTION_URL", "MQ_EXCHANGE_NAME",
    "MQ_EXCHANGE_TYPE", "MQ_ROUTING_BASE", "MQ_MAX_RECONNECT_DELAY",
    "publisher_create", "publisher_destroy", "publisher_start", "publisher_stop",
    "publisher_get_state", "publisher_get_state_msg", "publisher_publish",
    # subscriber_pomes
    "subscriber_create", "subscriber_destroy", "subscriber_start", "subscriber_stop",
    "subscriber_get_state", "subscriber_get_state_msg",
]

from importlib.metadata import version
__version__ = version("pypomes_messaging")
__version_info__ = tuple(int(i) for i in __version__.split(".") if i.isdigit())
